import io
import logging
import pathlib
import traceback
from contextlib import redirect_stdout

from .loaders import default_loader
from .bind import Bind


def _get_logger():
    return logging.getLogger(__name__)


def _print_loader_exception(e):
    msg = io.StringIO()
    with redirect_stdout(msg):
        traceback.print_exc()

    msg_lines = msg.getvalue().split('\n')
    msg_lines.append('{}: {}'.format(e.__class__.__name__, e))

    for line in msg_lines:
        _get_logger().info('LOADER: {}'.format(line))


def _action_name(name):
    return name + '_load'


class BindError(Exception):
    pass


def _get_path_property(name, explicit_bind):

    def _joined_path(self):
        path = getattr(self, _hidden_name(name))
        abs_path = self._root / str(path)

        if self._workspace is None:
            if explicit_bind:
                raise BindError(
                    "Missing workspace in PathStructure initializer. "
                    "Explicit bind for '{}' failed".format(name)
                )
            return abs_path

        action_name = _action_name(name)

        tracker = self._workspace.track(action_name)

        msg = "Action '{}' status: {}".format(
            action_name,
            tracker.status(),
        )
        _get_logger().info(msg)

        if tracker.finished():
            return abs_path

        if not tracker.started():
            _get_logger().info("Starting action: '{}'".format(action_name))
            if path.loader(abs_path, self._workspace.root, action_name):
                return abs_path
            else:
                return None
        else:
            _get_logger().info("Path not ready: '{}' in progress".format(action_name))
            return None

    return property(_joined_path)


def _hidden_name(name):
    return '_' + name


def _handle_attribute(name, value):
    if name.startswith('_'):
        return {name: value}

    if isinstance(value, Bind):
        loader = value.loader
        explicit_bind = True
    elif isinstance(value, str) or isinstance(value, pathlib.Path):
        loader = default_loader
        explicit_bind = False
    else:
        raise ValueError('PathStructure value type must be one of {}'.format(
            Bind, str, pathlib.Path,
        ))

    return {
        _hidden_name(name): Bind(str(value), loader),
        name: _get_path_property(name, explicit_bind),
    }


class PathStructureMeta(type):
    def __new__(mcs, cls_name, bases, attrib):
        mapping = {}

        for name, value in attrib.items():
            mapping.update(_handle_attribute(name, value))

        return super().__new__(mcs, cls_name, bases, mapping)



import os


def default_loader(target_path, workspace_root, action_name):
    from .workspace import Workspace

    ws = Workspace.construct(workspace_root)
    with ws.track(action_name):
        is_dir = '.' not in target_path.name

        if is_dir:
            os.makedirs(str(target_path))

    return True



from typing import Type, Union

from .base_path_structure import BasePathStructure
from .action_tracker import ActionTracker
from .base_action import BaseAction
from .state import WorkspaceState
from .base_data import BaseData
from .constants import _STATE


class Workspace:
    """Base class for Workspace subclasses

    Examples
    --------

    :ref:`basic-workspace`

    :ref:`workspace-with-chained-loaders`

    """

    @classmethod
    def construct(cls,
                  root,
                  path_structure_cls: Type[BasePathStructure] = None,
                  data_cls: Type[BaseData] = None, ):
        """Creates a workspace from scratch

        Parameters
        ----------
        path_structure_cls: BasePathStructure
            Filled path structure to create Workspace instances
        data_cls: BaseData
            Interface to automatically access data in the workspace
        """

        path = (
            path_structure_cls(root)
            if path_structure_cls
            else BasePathStructure(root)
        )
        data = data_cls() if data_cls else None
        return Workspace(path, data)

    def __init__(self, path, data=None):
        assert path is not None
        self.path = path
        self.data = data

        self.path._workspace = self
        self.state_path = self.root / _STATE
        self.state = WorkspaceState(self.state_path)

        if self.data:
            self.data._workspace = self


    @property
    def root(self):
        return self.path._root

    def __repr__(self):
        return '<{}({})>'.format(self.__class__.__name__, self.root)

    @property
    def id(self):
        """Workspace id attribute

        Id is used to sort workspaces from oldest to latest
        """
        dirname = self.root.name
        try:
            return int(dirname)
        except ValueError:
            return str(dirname)

    def track(self, action: str) -> ActionTracker:
        """Tracks the action inside the workspace

        Creates an ActionTracker object with action and this workspace as constructor parameters

        Parameters
        ----------
        action: subclass of BaseAction
            Action to track
        """
        return ActionTracker(action, self)

    def __eq__(self, other):
        return self.root == other.root
